from dataclasses import dataclass
from enum import IntEnum


class SnoozDeviceModel(IntEnum):
    UNSUPPORTED = 0
    ORIGINAL = 1
    PRO = 2
    BREEZ = 3


class SnoozFirmwareVersion(IntEnum):
    V2 = 2
    V3 = 3
    V4 = 4
    V5 = 5
    V6 = 6
    V7 = 7
    V8 = 8
    V9 = 9
    V10 = 10
    V11 = 11
    V12 = 12
    V13 = 13
    V14 = 14
    V15 = 15


@dataclass
class SnoozDeviceCharacteristicData:
    manufacturer: str
    model: str
    hardware: str
    firmware: str
    software: str | None


@dataclass
class SnoozAdvertisementData:
    model: SnoozDeviceModel
    firmware_version: SnoozFirmwareVersion
    password: str | None

    @property
    def is_pairing(self) -> bool:
        return self.password is not None

    @property
    def supports_fan(self) -> bool:
        return self.model == SnoozDeviceModel.BREEZ


@dataclass(repr=False)
class SnoozDeviceState:
    on: bool | None = None
    volume: int | None = None

    # Breez specific
    fan_on: bool | None = None
    fan_speed: int | None = None
    fan_auto_enabled: bool | None = None
    temperature: float | None = None
    target_temperature: int | None = None

    def __repr__(self) -> str:
        if self == UnknownSnoozState:
            return "Snooz(Unknown)"

        attributes = [f"Noise {'On' if self.on else 'Off'} at {self.volume}% volume"]
        fan_attrs: list[str] = []

        if self.fan_on is not None:
            fan_attrs += [f"Fan {'On' if self.fan_on else 'Off'}"]

        if self.fan_speed is not None:
            fan_attrs += [f"at {self.fan_speed}% speed"]

        if self.fan_auto_enabled is True:
            fan_attrs += ["[Auto]"]

        if len(fan_attrs) > 0:
            attributes += [" ".join(fan_attrs)]

        if self.temperature is not None:
            attributes += [f"{self.temperature}°F"]

        if self.target_temperature is not None:
            attributes += [f"{self.target_temperature}°F target"]

        parts = ", ".join(attributes)

        return f"Snooz({parts})"


UnknownSnoozState = SnoozDeviceState()



from __future__ import annotations

import asyncio
import logging
import struct
from asyncio import Lock
from dataclasses import dataclass
from enum import IntEnum
from typing import Callable

from bleak import BleakClient, BleakGATTCharacteristic
from bleak.exc import BleakDBusError
from events import Events

from .const import (
    FIRMWARE_REVISION_CHARACTERISTIC,
    HARDWARE_REVISION_CHARACTERISTIC,
    MANUFACTURER_NAME_CHARACTERISTIC,
    MODEL_NUMBER_CHARACTERISTIC,
    READ_COMMAND_CHARACTERISTIC,
    READ_STATE_CHARACTERISTIC,
    SOFTWARE_REVISION_CHARACTERISTIC,
    WRITE_STATE_CHARACTERISTIC,
)
from .model import SnoozDeviceCharacteristicData, SnoozDeviceState

# values less than this have no effect
MIN_DEVICE_VOLUME = 10
MIN_FAN_SPEED = 10

# When auto fan is enabled, the device will poll for fan power state updates since
# the device doesn't push them. This is the interval of temperature updates to
# wait for before requesting the current state of the device.
AUTO_FAN_TEMP_POLL_INTERVAL = 5

# number of times to retry a transient command failure before giving up
RETRY_WRITE_FAILURE_COUNT = 5
RETRY_SLEEP_DURATIONS = [0, 0.5, 1, 1, 2]
DBUS_ERRORS_TO_RETRY = (
    "org.bluez.Error",
    "org.bluez.Error.Failed",
    "org.bluez.Error.InProgress",
)

_LOGGER = logging.getLogger(__name__)


class Command(IntEnum):
    MOTOR_SPEED = 1
    MOTOR_ENABLED = 2
    TOP_LED = 3
    BOTTOM_LED = 4
    TURN_OFF_TIMER = 5
    PASSWORD = 6
    UPDATE_STATUS = 7
    TURN_ON_TIMER = 8
    SET_ON_TIMERS = 9
    SET_OFF_TIMERS = 10
    SYNC_TIME = 11
    SCHEDULE_REQUEST = 12
    NIGHTLIGHT = 13
    FADE_ON_TIME = 14
    FADE_OFF_TIME = 15
    REQUEST_FADE_TIMES = 16
    CAPBUTTON_POWER = 17
    CAPBUTTON_FASTER = 18
    CAPBUTTON_SLOWER = 19
    CAPBUTTON_NIGHTLIGHT = 20
    CAPBUTTON_DEBOUNCE = 21
    CAPBUTTON_SAMPLE_ARRAY = 22
    REQUEST_CAPBUTTON = 23
    CAPBUTTON_IIR_CO = 24
    SMARTPLUG_MODE = 25
    SET_FADE_SETTINGS = 26
    REQUEST_OTHER_SETTINGS = 27
    SET_TIMEZONE = 28
    SET_TIMEZONE_2 = 29
    SET_REAL_TIME = 30
    FAN_SPEED = 31
    FAN_ENABLED = 32
    AUTO_TEMP_ENABLED = 33
    AUTO_TEMP_THRESHOLD = 34
    ENABLE_LAST_POWERED_STATE = 35


class ResponseCommand(IntEnum):
    SEND_ON_SCHEDULE = 1
    SEND_OFF_SCHEDULE = 2
    SEND_FADE_SETTINGS = 3
    SEND_NIGHTLIGHT = 4
    SEND_CAPBUTTON = 5
    SEND_OTHER_SETTINGS = 6
    OTA_IDX = 7
    OTA_FAILED = 8
    TEMPERATURE = 9


class MissingCharacteristicError(Exception):
    """Raised when a required characteristic is missing on a client"""

    def __init__(self, uuid: str | None = None) -> None:
        super().__init__(f"Missing characteristic {uuid}")


@dataclass
class CharacteristicReference:
    uuid: str
    required: bool = True

    def get(self, client: BleakClient) -> BleakGATTCharacteristic:
        char = client.services.get_characteristic(self.uuid)

        if char is None and self.required:
            raise MissingCharacteristicError(self.uuid)

        return char


class SnoozDeviceApi:
    event_names = ("on_disconnect", "on_state_patched")

    def __init__(
        self,
        client: BleakClient | None = None,
        format_log_message: Callable[[str], str] | None = None,
    ) -> None:
        self.unsubscribe_all_events()
        self._client = client
        self._write_lock = Lock()
        self._ = format_log_message or (lambda msg: msg)
        self._read_state_char = CharacteristicReference(READ_STATE_CHARACTERISTIC)
        self._read_command_char = CharacteristicReference(READ_COMMAND_CHARACTERISTIC)
        self._write_state_char = CharacteristicReference(WRITE_STATE_CHARACTERISTIC)
        self._info_chars = [
            CharacteristicReference(MANUFACTURER_NAME_CHARACTERISTIC),
            CharacteristicReference(MODEL_NUMBER_CHARACTERISTIC),
            CharacteristicReference(HARDWARE_REVISION_CHARACTERISTIC),
            CharacteristicReference(FIRMWARE_REVISION_CHARACTERISTIC),
            # Older Snooz devices don't have this
            CharacteristicReference(SOFTWARE_REVISION_CHARACTERISTIC, required=False),
        ]
        self._required_chars = [
            *[
                ref
                for ref in [getattr(self, c) for c in dir(self)]
                if isinstance(ref, CharacteristicReference) and ref.required
            ],
            *[ref for ref in self._info_chars if ref.required],
        ]

    @property
    def is_connected(self) -> bool:
        return self._client is not None and self._client.is_connected

    def load_client(self, client: BleakClient) -> None:
        self._client = client
        self._info = None

        for char in self._required_chars:
            char.get(client)

    def _require_client(self, require_connection: bool = True) -> BleakClient:
        if self._client is None:
            raise AssertionError("client was None")

        if require_connection and not self._client.is_connected:
            raise AssertionError("client was not connected")

        return self._client

    def unsubscribe_all_events(self) -> None:
        self.events = Events(self.event_names)

    async def async_disconnect(self) -> None:
        client = self._require_client()

        await client.disconnect()

    async def async_get_info(self) -> SnoozDeviceCharacteristicData | None:
        client = self._require_client(False)
        if not client.is_connected:
            return None
        string_props: list[str | None] = []
        for char_ref in self._info_chars:
            char = char_ref.get(client)
            if not char:
                value = None
            else:
                value = (await client.read_gatt_char(char)).decode().split("\0")[0]
            string_props.append(value)

        return SnoozDeviceCharacteristicData(*string_props)  # type: ignore

    async def async_authenticate_connection(self, password: str) -> None:
        await self._async_write_state(
            bytes([Command.PASSWORD, *bytes.fromhex(password)])
        )

    async def async_set_power(self, on: bool) -> None:
        await self._async_write_state(bytes([Command.MOTOR_ENABLED, 1 if on else 0]))

    async def async_set_fan_power(self, on: bool) -> None:
        await self._async_write_state(bytes([Command.FAN_ENABLED, 1 if on else 0]))

    async def async_set_auto_temp_enabled(self, on: bool) -> None:
        await self._async_write_state(
            bytes([Command.AUTO_TEMP_ENABLED, 1 if on else 0])
        )
        await self.async_request_other_settings()

    async def async_set_volume(self, volume: int) -> None:
        if volume < 0 or volume > 100:
            raise ValueError(f"Volume must be between 0 and 100 - got {volume}")

        await self._async_write_state(bytes([Command.MOTOR_SPEED, volume]))

    async def async_set_fan_speed(self, speed: int) -> None:
        if speed < 0 or speed > 100:
            raise ValueError(f"Speed must be between 0 and 100 - got {speed}")

        await self._async_write_state(bytes([Command.FAN_SPEED, speed]))

    async def async_set_auto_temp_threshold(self, threshold_f: int) -> None:
        if threshold_f < 0 or threshold_f > 100:
            raise ValueError(
                f"Temperature must be between 0 and 100 - got {threshold_f}"
            )

        await self._async_write_state(bytes([Command.AUTO_TEMP_THRESHOLD, threshold_f]))
        await self.async_request_other_settings()

    async def async_read_state(self, use_cached: bool = False) -> SnoozDeviceState:
        client = self._require_client()

        data = await client.read_gatt_char(
            self._read_state_char.get(client), use_cached=use_cached
        )
        return unpack_state(data)

    async def async_subscribe(self) -> None:
        client = self._require_client(False)

        if not client.is_connected:
            return

        def on_state_change(_: BleakClient, payload: bytes) -> None:
            self.events.on_state_patched(unpack_state(payload))

        await client.start_notify(
            self._read_state_char.get(client),
            on_state_change,
        )

        def on_response_command(_: BleakClient, data: bytes) -> None:
            command = ResponseCommand(data[0])
            payload = data[1:]

            self.events.on_state_patched(unpack_response_command(command, payload))

        await client.start_notify(
            self._read_command_char.get(client), on_response_command
        )

    async def async_request_other_settings(self) -> None:
        await self._async_write_state(bytes([Command.REQUEST_OTHER_SETTINGS]))

    async def _async_write_state(self, data: bytes) -> None:
        client = self._require_client(False)

        attempts = 0

        async with self._write_lock:
            last_ex: BleakDBusError | None = None

            while client.is_connected and attempts <= RETRY_WRITE_FAILURE_COUNT:
                try:
                    message = f"write {data.hex('-')}"
                    if attempts > 0 and last_ex is not None:
                        message += f" (attempt {attempts+1}, last error: {last_ex})"
                    _LOGGER.debug(self._(message))
                    await client.write_gatt_char(
                        self._write_state_char.get(client), data, response=True
                    )
                    return
                except BleakDBusError as ex:
                    last_ex = ex
                    if ex.dbus_error in DBUS_ERRORS_TO_RETRY:
                        sleep_duration = RETRY_SLEEP_DURATIONS[
                            attempts % len(RETRY_SLEEP_DURATIONS)
                        ]
                        attempts += 1

                        if attempts > RETRY_WRITE_FAILURE_COUNT:
                            raise Exception(
                                f"Got transient error {attempts} times"
                            ) from ex

                        await asyncio.sleep(sleep_duration)
                    else:
                        raise


def unpack_response_command(command: ResponseCommand, data: bytes) -> SnoozDeviceState:
    result = SnoozDeviceState()

    if command == ResponseCommand.SEND_OTHER_SETTINGS:
        (auto_enabled, target_temperature) = struct.unpack("<xxxxxxxxxxBB", data[0:12])

        result.fan_auto_enabled = bool(auto_enabled)
        result.target_temperature = target_temperature
    if command == ResponseCommand.TEMPERATURE:
        [temp] = struct.unpack("<f", data[0:4])
        result.temperature = round(temp, 2)

    return result


def unpack_state(data: bytes) -> SnoozDeviceState:
    (volume, on, fan_speed, fan_on) = struct.unpack("<BBxBB", data[0:5])

    return SnoozDeviceState(
        volume=volume,
        on=bool(on),
        fan_on=bool(fan_on),
        fan_speed=fan_speed,
    )



from pysnooz.model import SnoozAdvertisementData, SnoozDeviceState

SOUND_PROPS = ["on", "volume"]
FAN_PROPS = [
    "fan_on",
    "fan_speed",
    "fan_auto_enabled",
    "temperature",
    "target_temperature",
]


class SnoozStateStore:
    def __init__(self, adv_data: SnoozAdvertisementData) -> None:
        self.current = SnoozDeviceState()
        self._adv_data = adv_data

    def patch(self, state: SnoozDeviceState) -> bool:
        updated_props = [
            p
            for p in dir(state)
            if not p.startswith("__")
            and getattr(state, p) is not None
            and self._supports_prop(p)
        ]
        did_change = False
        for prop in updated_props:
            p_current = getattr(self.current, prop)
            p_state = getattr(state, prop)

            if p_current != p_state:
                did_change = True
                setattr(self.current, prop, p_state)

        return did_change

    def _supports_prop(self, prop: str) -> bool:
        if not self._adv_data.supports_fan and prop in FAN_PROPS:
            return False

        return True



from __future__ import annotations

import logging
from dataclasses import dataclass

from home_assistant_bluetooth import BluetoothServiceInfo

from pysnooz.const import (
    FIRMWARE_PAIRING_FLAGS,
    FIRMWARE_VERSION_BY_FLAGS,
    MODEL_NAME_BREEZ,
    SNOOZ_ADVERTISEMENT_LENGTH,
    SUPPORTED_MODEL_NAMES,
)
from pysnooz.model import SnoozAdvertisementData, SnoozDeviceModel, SnoozFirmwareVersion

_LOGGER = logging.getLogger(__name__)


@dataclass
class ParsedAdvertisementFlags:
    firmware_version: SnoozFirmwareVersion
    is_pairing: bool


def parse_firmware_flags(flags: int) -> ParsedAdvertisementFlags | None:
    is_pairing = (FIRMWARE_PAIRING_FLAGS & flags) == FIRMWARE_PAIRING_FLAGS
    flags_without_pairing = flags & ~FIRMWARE_PAIRING_FLAGS

    if flags_without_pairing not in FIRMWARE_VERSION_BY_FLAGS:
        _LOGGER.debug(f"Unknown device flags {flags_without_pairing:02X}")
        return None

    return ParsedAdvertisementFlags(
        FIRMWARE_VERSION_BY_FLAGS[flags_without_pairing], is_pairing
    )


def parse_snooz_advertisement(
    data: BluetoothServiceInfo,
) -> SnoozAdvertisementData | None:
    advertisement = data.manufacturer_data.get(data.manufacturer_id)

    if advertisement is None:
        return None

    if len(advertisement) != SNOOZ_ADVERTISEMENT_LENGTH:
        return None

    if not (flags := parse_firmware_flags(advertisement[0])):
        return None

    model = get_device_model(data.name, flags.firmware_version)

    if model == SnoozDeviceModel.UNSUPPORTED:
        _LOGGER.debug(
            f"{data.name} is unsupported with firmware {flags.firmware_version}"
        )
        return None

    return SnoozAdvertisementData(
        model,
        flags.firmware_version,
        password=advertisement[1:].hex() if flags.is_pairing else None,
    )


def match_known_device_name(advertised_name: str) -> str | None:
    return next(
        (
            value
            for value in SUPPORTED_MODEL_NAMES
            if advertised_name.lower().startswith(value.lower())
        ),
        None,
    )


def get_device_model(
    advertised_name: str, firmware: SnoozFirmwareVersion
) -> SnoozDeviceModel:
    known = match_known_device_name(advertised_name)

    # The advertised name is in the format "Snooz-XXXX" or "Breez-XXXX"
    # we require this format since the name stored in device characteristics
    # is always "Snooz" and matching it would not correctly identify Breez models
    if known is None or len(advertised_name) != len(known) + 5:
        return SnoozDeviceModel.UNSUPPORTED

    if firmware in [
        SnoozFirmwareVersion.V2,
        SnoozFirmwareVersion.V3,
        SnoozFirmwareVersion.V4,
        SnoozFirmwareVersion.V5,
    ]:
        return SnoozDeviceModel.ORIGINAL

    if firmware == SnoozFirmwareVersion.V6:
        return (
            SnoozDeviceModel.BREEZ
            if known == MODEL_NAME_BREEZ
            else SnoozDeviceModel.PRO
        )

    return SnoozDeviceModel.UNSUPPORTED


def get_device_display_name(advertised_name: str, address: str) -> str:
    known = match_known_device_name(advertised_name)
    if known is not None:
        return f"{known} {address.replace(':', '')[-4:]}"

    return advertised_name.replace("-", " ")



from __future__ import annotations

import asyncio
import logging
from asyncio import AbstractEventLoop, CancelledError, Event, Lock, Task
from datetime import datetime
from enum import Enum
from typing import Any, Callable

from bleak.backends.device import BLEDevice
from bleak_retry_connector import (
    BleakAbortedError,
    BleakClientWithServiceCache,
    BleakConnectionError,
    BleakNotFoundError,
    establish_connection,
)
from events import Events
from transitions import EventData, Machine, State

from pysnooz.advertisement import get_device_display_name
from pysnooz.api import MissingCharacteristicError, SnoozDeviceApi
from pysnooz.commands import (
    CommandProcessorState,
    SnoozCommandData,
    SnoozCommandProcessor,
    SnoozCommandResult,
    create_command_processor,
    get_device_info,
)
from pysnooz.const import UNEXPECTED_ERROR_LOG_MESSAGE
from pysnooz.model import (
    SnoozAdvertisementData,
    SnoozDeviceCharacteristicData,
    SnoozDeviceModel,
    SnoozDeviceState,
    SnoozFirmwareVersion,
    UnknownSnoozState,
)
from pysnooz.store import SnoozStateStore

_LOGGER = logging.getLogger(__name__)

MAX_RECONNECTION_ATTEMPTS = 3
RECONNECTION_DELAY_SECONDS = 0.25
DEVICE_UNAVAILABLE_EXCEPTIONS = (
    BleakAbortedError,
    BleakConnectionError,
    BleakNotFoundError,
    MissingCharacteristicError,
)


class SnoozConnectionStatus(Enum):
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2


class DisconnectionReason(Enum):
    # disconnection was initiated by the user
    USER = 0
    # the bluetooth connection was lost
    DEVICE = 1
    # a connection attempt failed with a known error
    # like a timeout or device not found or missing characteristic
    DEVICE_UNAVAILABLE = 2
    # an exception was thrown during connection or
    # command execution that was uncaught
    UNEXPECTED_ERROR = 3


class SnoozDeviceUnavailableError(Exception):
    pass


class SnoozDevice:
    def __init__(
        self,
        device: BLEDevice,
        adv_data: SnoozAdvertisementData,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self.events = Events(
            (
                # one or more state properties have changed
                "on_state_change",
                # SnoozConnectionStatus
                "on_connection_status_change",
                # time it takes to create a bluetooth connection, authenticate
                # and subscribe to notifications
                "on_connection_load_time",
                # time between connection ready and disconnection
                "on_connection_duration",
            )
        )
        self._device = device
        self._adv_data = adv_data
        self._loop = loop if loop is not None else asyncio.get_running_loop()
        self._last_dispatched_connection_status: SnoozConnectionStatus | None = None
        self._connection_complete = Event()
        self._connections_exhausted = Event()
        self._connection_attempts: int = 0
        self._connection_start_time: datetime | None = None
        self._connection_ready_time: datetime | None = None
        self._api: SnoozDeviceApi | None = None
        self._store = SnoozStateStore(self._adv_data)
        self._connect_lock = Lock()
        self._command_lock = Lock()
        self._connection_task: Task[None] | None = None
        self._reconnection_task: Task[None] | None = None
        self._current_command: SnoozCommandProcessor | None = None
        self._expected_disconnect: bool = False
        self._last_disconnect_reason: DisconnectionReason | None = None

        not_disconnected = [
            SnoozConnectionStatus.CONNECTING,
            SnoozConnectionStatus.CONNECTED,
        ]

        states = [
            State(
                SnoozConnectionStatus.DISCONNECTED,
                on_enter=self._on_device_disconnected,
            ),
            State(SnoozConnectionStatus.CONNECTING),
            State(SnoozConnectionStatus.CONNECTED, on_enter=self._on_connection_ready),
        ]

        self._machine = Machine(
            model_attribute="connection_status",
            states=states,
            initial=SnoozConnectionStatus.DISCONNECTED,
            after_state_change=self._on_connection_status_change,
            send_event=True,
            on_exception=self._on_machine_exception,
        )
        self._machine.add_transition(
            "connection_start",
            SnoozConnectionStatus.DISCONNECTED,
            SnoozConnectionStatus.CONNECTING,
            before=self._on_connection_start,
        )
        self._machine.add_transition(
            "connection_ready",
            SnoozConnectionStatus.CONNECTING,
            SnoozConnectionStatus.CONNECTED,
        )
        self._machine.add_transition(
            "device_disconnected",
            not_disconnected,
            SnoozConnectionStatus.DISCONNECTED,
            after=self._after_device_disconnected,
        )
        self._machine.add_transition(
            "command_execution_cancelled",
            [SnoozConnectionStatus.DISCONNECTED, SnoozConnectionStatus.CONNECTED],
            None,
            before=lambda _: self._cancel_current_command(),
        )

        def set_expected_disconnect(expected: bool) -> None:
            self._expected_disconnect = expected

        self._machine.add_transition(
            "command_execution_cancelled",
            SnoozConnectionStatus.CONNECTING,
            SnoozConnectionStatus.DISCONNECTED,
            prepare=lambda _: set_expected_disconnect(True),
            before=lambda _: self._cancel_current_command(),
            after=lambda _: set_expected_disconnect(False),
        )

        self.events.on_connection_load_time += lambda t: _LOGGER.debug(
            self._(f"Connection load time: {t}")
        )
        self.events.on_connection_duration += lambda t: _LOGGER.debug(
            self._(f"Connection duration: {t}")
        )

    @property
    def name(self) -> str:
        return self._device.name

    @property
    def address(self) -> str:
        return self._device.address

    @property
    def model(self) -> SnoozDeviceModel:
        return self._adv_data.model

    @property
    def firmware_version(self) -> SnoozFirmwareVersion:
        return self._adv_data.firmware_version

    @property
    def connection_status(self) -> SnoozConnectionStatus:
        return self._machine.connection_status

    @property
    def is_connected(self) -> bool:
        return self.connection_status == SnoozConnectionStatus.CONNECTED

    @property
    def state(self) -> SnoozDeviceState:
        return self._store.current

    def subscribe_to_state_change(
        self, callback: Callable[[], None]
    ) -> Callable[[], None]:
        """
        Subscribe to device state and connection status change.
        Returns a callback to unsubscribe.
        """

        def wrapped_callback(*_: Any) -> None:
            callback()

        self.events.on_state_change += wrapped_callback
        self.events.on_connection_status_change += wrapped_callback

        def unsubscribe() -> None:
            self.events.on_state_change -= wrapped_callback
            self.events.on_connection_status_change -= wrapped_callback

        return unsubscribe

    async def async_disconnect(self) -> None:
        if self.connection_status == SnoozConnectionStatus.DISCONNECTED:
            return

        self._expected_disconnect = True
        try:
            self._cancel_current_command()

            if (
                self._reconnection_task is not None
                and not self._reconnection_task.done()
            ):
                self._reconnection_task.cancel()

            if self._connection_task is not None and not self._connection_task.done():
                self._connection_task.cancel()

            if self._api is not None:
                await self._api.async_disconnect()

            self._machine.device_disconnected(reason=DisconnectionReason.USER)
        finally:
            self._expected_disconnect = False

    async def async_get_info(self) -> SnoozDeviceCharacteristicData | None:
        result = await self.async_execute_command(get_device_info())

        return result.response

    async def async_execute_command(self, data: SnoozCommandData) -> SnoozCommandResult:
        self._cancel_current_command()

        start_time = datetime.now()
        command = create_command_processor(
            self._loop,
            start_time,
            data,
            format_log_message=self._,
        )
        self._current_command = command

        try:
            await self._async_execute_current_command()
        except CancelledError:
            self._machine.command_execution_cancelled()

            # when async_disconnect() is called, return a cancelled command
            # instead of raising since the cancellation is expected by the user
            if (
                not self.is_connected
                and self._last_disconnect_reason == DisconnectionReason.USER
            ):
                pass
            else:
                raise

        result = await command.result

        self._current_command = None

        return result

    def _cancel_current_command(self) -> None:
        if self._current_command is None:
            return

        self._current_command.cancel()
        self._current_command = None

    async def _async_execute_current_command(self) -> None:
        command = self._current_command

        try:
            try:
                await self._async_wait_for_connection_complete()
            except CancelledError:
                raise

            async with self._command_lock:
                if (
                    self._api is not None
                    and command is not None
                    and command.state != CommandProcessorState.COMPLETE
                ):
                    await command.async_execute(self._api, raise_on_cancel=True)
        except SnoozDeviceUnavailableError:
            self._machine.device_disconnected(
                reason=DisconnectionReason.DEVICE_UNAVAILABLE
            )
        except Exception:
            self._machine.device_disconnected(
                reason=DisconnectionReason.UNEXPECTED_ERROR
            )

    async def _async_wait_for_connection_complete(self) -> None:
        if self.connection_status == SnoozConnectionStatus.DISCONNECTED:
            async with self._connect_lock:
                if self._connection_task is None or self._connection_task.done():
                    self._connection_task = self._loop.create_task(
                        self._async_connect(), name=f"[Connect] {self.display_name}"
                    )
            await self._connection_task

        await self._connection_complete.wait()

    async def _async_connect(self) -> None:
        self._machine.connection_start()

        try:
            api = await self._async_create_api()
            api.events.on_disconnect += lambda: self._machine.device_disconnected(
                reason=DisconnectionReason.DEVICE
            )

            def apply_state_patch(state: SnoozDeviceState) -> None:
                if self._store.patch(state):
                    self.events.on_state_change(self.state)

            api.events.on_state_patched += apply_state_patch

            self._before_device_connected()
            self._api = api
        except DEVICE_UNAVAILABLE_EXCEPTIONS as ex:
            raise SnoozDeviceUnavailableError() from ex

        # ensure each call with side effects checks the connection status
        # to prevent a cancellation race condition

        if self.connection_status == SnoozConnectionStatus.CONNECTING:
            if self._adv_data.password is None:
                raise ValueError("Missing device password")
            await api.async_authenticate_connection(self._adv_data.password)

        if self.connection_status == SnoozConnectionStatus.CONNECTING:
            await api.async_subscribe()

        if self.connection_status == SnoozConnectionStatus.CONNECTING:
            self._machine.connection_ready()

    async def _async_create_api(self) -> SnoozDeviceApi:
        api = SnoozDeviceApi(format_log_message=self._)

        def _on_disconnect(_: BleakClientWithServiceCache) -> None:
            # don't trigger a device disconnection event when a user
            # manually requests a disconnect or we're reconnecting
            if not self._expected_disconnect:
                api.events.on_disconnect()

        client = await establish_connection(
            BleakClientWithServiceCache,
            self._device,
            self.display_name,
            _on_disconnect,
            use_services_cache=True,
            ble_device_callback=lambda: self._device,
        )

        try:
            api.load_client(client)
        except MissingCharacteristicError:
            await client.clear_cache()

            self._expected_disconnect = True
            try:
                await client.disconnect()
            finally:
                self._expected_disconnect = False

            raise

        return api

    def _cleanup_api(self) -> None:
        if self._api is not None:
            self._api.unsubscribe_all_events()
        self._api = None

    async def _async_reconnect(self) -> None:
        await asyncio.sleep(RECONNECTION_DELAY_SECONDS)
        await self._async_execute_current_command()

    def _on_machine_exception(self, e: EventData) -> None:
        # make sure any pending commands are completed
        if (
            self._current_command is not None
            and self._current_command.state != CommandProcessorState.COMPLETE
        ):
            self._current_command.on_unhandled_exception()

        _LOGGER.exception(
            self._(
                "An exception occurred during a state transition.\n"
                + UNEXPECTED_ERROR_LOG_MESSAGE
            )
        )

    def _on_connection_start(self, e: EventData) -> None:
        message = "Start connection"
        if self._connection_attempts >= 1:
            message += f" (attempt {self._connection_attempts})"
        _LOGGER.debug(self._(message))

        self._connection_complete.clear()
        self._connections_exhausted.clear()
        self._connection_attempts += 1
        self._connection_start_time = datetime.now()
        self._last_disconnect_reason = None

    def _before_device_connected(self) -> None:
        start_time = datetime.now()
        if self._connection_start_time is not None:
            start_time = self._connection_start_time
        message = f"Got connection in {datetime.now() - start_time}"
        if self._connection_attempts >= 1:
            message += f" (attempt {self._connection_attempts})"
        _LOGGER.debug(self._(message))

    def _on_connection_ready(self, e: EventData) -> None:
        self._connection_attempts = 0
        self._connection_ready_time = datetime.now()
        if self._connection_start_time is not None:
            self.events.on_connection_load_time(
                self._connection_ready_time - self._connection_start_time
            )
        self._connection_complete.set()
        self._connections_exhausted.clear()

    def _on_connection_status_change(self, e: EventData) -> None:
        new_status = self.connection_status

        if new_status == self._last_dispatched_connection_status:
            return

        _LOGGER.debug(self._(describe_connection_status(new_status)))

        self._last_dispatched_connection_status = new_status
        self.events.on_connection_status_change(new_status)

    def _after_device_disconnected(self, e: EventData) -> None:
        reason: DisconnectionReason = e.kwargs.get("reason")
        level = (
            logging.ERROR
            if reason == DisconnectionReason.DEVICE_UNAVAILABLE
            else logging.INFO
        )
        _LOGGER.log(
            level,
            self._(f"disconnected because {describe_disconnection_reason(reason)}"),
            exc_info=reason
            not in (DisconnectionReason.USER, DisconnectionReason.DEVICE),
        )

    def _on_device_disconnected(self, e: EventData) -> None:
        self._cleanup_api()

        last_event = self._connection_ready_time or self._connection_start_time
        if last_event is not None:
            self.events.on_connection_duration(datetime.now() - last_event)

        self._connection_start_time = None
        self._connection_ready_time = None
        self._connection_complete.set()

        disconnect_reason: DisconnectionReason = e.kwargs.get("reason")
        self._last_disconnect_reason = disconnect_reason

        # if the disconnection was initiated from the user, don't attempt to reconnect
        if disconnect_reason == DisconnectionReason.USER or self._expected_disconnect:
            return

        if self._connection_attempts >= MAX_RECONNECTION_ATTEMPTS:
            _LOGGER.error(
                self._(
                    f"Unavailable after {self._connection_attempts}"
                    " connection attempts"
                )
            )

            if (
                self._current_command is not None
                and self._current_command.state != CommandProcessorState.COMPLETE
            ):
                self._current_command.on_device_unavailable()

            self._connections_exhausted.set()
            self._connection_attempts = 0

            return

        if self._current_command is not None:
            # don't reconnect on unexpected errors
            if disconnect_reason == DisconnectionReason.UNEXPECTED_ERROR:
                self._current_command.on_unhandled_exception()
                return

            self._current_command.on_disconnected()

        # attempt to reconnect automatically
        # we don't await the result since this is called from a sync state transition
        # we cleanup this task on disconnect
        reconnection_task = self._loop.create_task(
            self._async_reconnect(),
            name=f"[Reconnect] {self.display_name}",
        )

        # cancel previous tasks to avoid any zombies
        if self._reconnection_task is not None and not self._reconnection_task.done():
            self._reconnection_task.cancel()

        self._reconnection_task = reconnection_task

    def _(self, message: str) -> str:
        """Format a message for logging."""
        return f"[{self.display_name}] {message}"

    @property
    def display_name(self) -> str:
        return get_device_display_name(self.name, self.address)

    def __repr__(self) -> str:
        description: list[str] = []

        if self.is_connected:
            state = self.state
            if state is None or state == UnknownSnoozState:
                description += ["Unknown state"]
            else:
                description += [state.__repr__()]
        else:
            description += ["Disconnected"]

        return f"SnoozDevice({self.address} {' '.join(description)})"


def describe_connection_status(status: SnoozConnectionStatus) -> str:
    descriptions = {
        SnoozConnectionStatus.DISCONNECTED: "🔴 Disconnected",
        SnoozConnectionStatus.CONNECTING: "🟡 Connecting",
        SnoozConnectionStatus.CONNECTED: "🟢 Connected",
    }
    return descriptions[status] or status.name


def describe_disconnection_reason(reason: DisconnectionReason) -> str:
    descriptions = {
        DisconnectionReason.USER: (
            f"{SnoozDevice.async_disconnect.__qualname__}() was called"
        ),
        DisconnectionReason.DEVICE: "the bluetooth connection was lost",
        DisconnectionReason.UNEXPECTED_ERROR: (
            f"an uncaught exception occurred.\n{UNEXPECTED_ERROR_LOG_MESSAGE}"
        ),
        DisconnectionReason.DEVICE_UNAVAILABLE: (
            "the device couldn't establish a connection"
        ),
    }
    return descriptions[reason] or reason.name



"""Utilities for integration testing Snooz devices."""

from __future__ import annotations

import dataclasses
import logging
import struct
from typing import Any, Awaitable, Callable
from unittest.mock import MagicMock

from bleak import BleakClient, BLEDevice
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.service import BleakGATTServiceCollection
from bleak_retry_connector import BleakClientWithServiceCache

from pysnooz.api import (
    READ_COMMAND_CHARACTERISTIC,
    READ_STATE_CHARACTERISTIC,
    WRITE_STATE_CHARACTERISTIC,
    Command,
    ResponseCommand,
    SnoozDeviceApi,
)
from pysnooz.const import (
    FIRMWARE_REVISION_CHARACTERISTIC,
    HARDWARE_REVISION_CHARACTERISTIC,
    MANUFACTURER_NAME_CHARACTERISTIC,
    MODEL_NUMBER_CHARACTERISTIC,
    SOFTWARE_REVISION_CHARACTERISTIC,
)
from pysnooz.device import SnoozDevice
from pysnooz.model import SnoozAdvertisementData, SnoozDeviceModel, SnoozDeviceState

_LOGGER = logging.getLogger(__name__)


class MockSnoozDevice(SnoozDevice):
    """Used for testing integration with Bleak."""

    def __init__(
        self,
        address_or_ble_device: BLEDevice | str,
        adv_data: SnoozAdvertisementData,
        initial_state: SnoozDeviceState = SnoozDeviceState(),
    ) -> None:
        """Create a mock snooz device that does not make any bluetooth calls."""
        super().__init__(address_or_ble_device, adv_data)

        def _on_disconnected(_: BleakClient) -> None:
            if self._api is not None and not self._expected_disconnect:
                self._api.events.on_disconnect()

        state_copy = dataclasses.replace(initial_state)
        self._store.current = state_copy
        self._mock_client = MockSnoozClient(
            address_or_ble_device, adv_data.model, _on_disconnected
        )
        self._mock_client.trigger_state(state_copy)

        async def _create_mock_api() -> SnoozDeviceApi:
            self._mock_client.reset_mock()
            return SnoozDeviceApi(self._mock_client)

        self._async_create_api = _create_mock_api  # type: ignore

    def trigger_disconnect(self) -> None:
        """Trigger a disconnect."""
        if self._api is not None:
            self._mock_client.trigger_disconnect()

    def trigger_state(self, new_state: SnoozDeviceState) -> None:
        """Trigger a new state."""
        if self._api is not None:
            self._mock_client.trigger_state(new_state)

    def trigger_temperature(self, temp: float) -> None:
        """Trigger a new temperature update."""
        if self._api is not None:
            self._mock_client.trigger_temperature(temp)


CharNotifyCallback = Callable[
    [BleakGATTCharacteristic, bytearray], None | Awaitable[None]
]


class MockSnoozClient(BleakClientWithServiceCache):
    """Used for testing integration with Bleak."""

    def __init__(  # pylint: disable=super-init-not-called
        self,
        address_or_ble_device: BLEDevice | str,
        model: SnoozDeviceModel,
        disconnected_callback: Callable[[BleakClient], None] | None = None,
        **kwargs: Any,
    ):
        self._is_connected = True
        self._state = SnoozDeviceState(on=False, volume=10)
        if model == SnoozDeviceModel.BREEZ:
            self._state.fan_on = False
            self._state.fan_speed = 10
        self._disconnected_callback = disconnected_callback
        self._model = model

        self._has_set_password = False
        self._state_char_callback: CharNotifyCallback | None = None
        self._command_char_callback: CharNotifyCallback | None = None
        self._services = MagicMock(spec=BleakGATTServiceCollection)

        def mock_char(uuid: str) -> BleakGATTCharacteristic:
            if uuid not in [
                MODEL_NUMBER_CHARACTERISTIC,
                FIRMWARE_REVISION_CHARACTERISTIC,
                HARDWARE_REVISION_CHARACTERISTIC,
                SOFTWARE_REVISION_CHARACTERISTIC,
                MANUFACTURER_NAME_CHARACTERISTIC,
                READ_STATE_CHARACTERISTIC,
                WRITE_STATE_CHARACTERISTIC,
                READ_COMMAND_CHARACTERISTIC,
            ]:
                raise Exception(f"Unexpected char uuid: {uuid}")

            if (
                uuid == SOFTWARE_REVISION_CHARACTERISTIC
                and self._model == SnoozDeviceModel.ORIGINAL
            ):
                return None

            return MagicMock(spec=BleakGATTCharacteristic, uuid=uuid)

        self._services.get_characteristic.side_effect = mock_char

    @property
    def services(self) -> BleakGATTServiceCollection:
        return self._services

    async def connect(self, **kwargs: Any) -> bool:
        self._is_connected = True
        return True

    async def disconnect(self) -> bool:
        self.trigger_disconnect()

        return True

    def trigger_disconnect(self) -> None:
        """Set the device as disconnected and trigger callbacks."""
        if not self._is_connected:
            return

        _LOGGER.debug("Triggering device disconnect")

        self._is_connected = False

        if self._disconnected_callback is not None:
            self._disconnected_callback(self)

    def trigger_state(self, state: SnoozDeviceState) -> None:
        """Set the current state and notify subscribers."""
        _LOGGER.debug(f"Triggering state update {state}")

        self._state = dataclasses.replace(state)
        self._send_state_update()

    def trigger_temperature(self, temp: float) -> None:
        """Trigger a temperature update and notify subscribers."""
        _LOGGER.debug("Triggering temperature update {temp}")

        self._send_response_command(
            ResponseCommand.TEMPERATURE, struct.pack("<f", temp)
        )

    def reset_mock(self, initial_state: bool = False) -> None:
        """Reset the mock state."""
        _LOGGER.debug("Resetting mock client")
        self._is_connected = True
        self._has_set_password = False
        self._state_char_callback = None
        self._command_char_callback = None
        self._services.reset_mock()

        if initial_state:
            self._state = SnoozDeviceState(on=False, volume=10)

            if self._model == SnoozDeviceModel.BREEZ:
                self._state.fan_on = False
                self._state.fan_speed = 10

    async def pair(self, *args: Any, **kwargs: Any) -> bool:
        raise NotImplementedError()

    async def unpair(self) -> bool:
        raise NotImplementedError()

    @property
    def mtu_size(self) -> int:
        raise NotImplementedError()

    @property
    def is_connected(self) -> bool:
        return self._is_connected

    async def get_services(self, **kwargs: Any) -> BleakGATTServiceCollection:
        return self._services

    async def read_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic,
        **kwargs: Any,
    ) -> bytearray:
        if char_specifier.uuid == READ_STATE_CHARACTERISTIC:
            return self._get_state_char_data()

        mock_values = {
            **CHAR_VALUES_BY_MODEL[self._model],
            MANUFACTURER_NAME_CHARACTERISTIC: "Snooz",
        }

        return bytearray(mock_values[char_specifier.uuid], "utf-8")

    async def read_gatt_descriptor(self, handle: int, **kwargs: Any) -> bytearray:
        raise NotImplementedError()

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic,
        data: bytes | bytearray | memoryview,
        response: bool = False,
    ) -> None:
        if char_specifier.uuid != WRITE_STATE_CHARACTERISTIC:
            raise Exception(f"Unexpected char specifier: {char_specifier.uuid}")

        command = data[0]

        if command == Command.PASSWORD:
            self._has_set_password = True
            _LOGGER.debug(f"Received password: {data[1:].hex()}")
            return

        if self._has_set_password:
            if command == Command.REQUEST_OTHER_SETTINGS:
                self._send_response_command(
                    ResponseCommand.SEND_OTHER_SETTINGS,
                    pack_other_settings(self._state),
                )

                return
            elif command == Command.MOTOR_ENABLED:
                self._state.on = unpack_bool(data[1])
            elif command == Command.MOTOR_SPEED:
                self._state.volume = max(0, min(100, int(data[1])))
            elif command == Command.FAN_ENABLED:
                self._state.fan_on = unpack_bool(data[1])
            elif command == Command.FAN_SPEED:
                self._state.fan_speed = max(0, min(100, int(data[1])))
            elif command == Command.AUTO_TEMP_ENABLED:
                self._state.fan_auto_enabled = unpack_bool(data[1])
            elif command == Command.AUTO_TEMP_THRESHOLD:
                self._state.target_temperature = max(0, data[1])
            else:
                raise Exception(f"Unexpected command ID: {command} in {data.hex('-')}")
        else:
            _LOGGER.warning(
                f"Received command before password was set: {data.hex('-')}"
            )

        self._send_state_update()

    async def write_gatt_descriptor(
        self, handle: int, data: bytes | bytearray | memoryview
    ) -> None:
        raise NotImplementedError()

    async def start_notify(
        self,
        char_specifier: BleakGATTCharacteristic,
        callback: CharNotifyCallback | None,
        **kwargs: Any,
    ) -> None:
        uuid = char_specifier.uuid

        if uuid == READ_STATE_CHARACTERISTIC:
            self._state_char_callback = callback
        elif uuid == READ_COMMAND_CHARACTERISTIC:
            self._command_char_callback = callback
        else:
            raise Exception(f"Unexpected notification characteristic: {uuid}")

    async def stop_notify(self, char_specifier: BleakGATTCharacteristic) -> None:
        await self.start_notify(char_specifier, None)

    def _send_state_update(self) -> None:
        if self._state_char_callback is None:
            return

        # pass None since it's unused by SnoozDeviceApi
        self._state_char_callback(
            None,
            self._get_state_char_data() if self._has_set_password
            # when a password isn't set, the device sends a zeroed out state
            else bytearray([0] * 20),
        )

    def _send_response_command(self, command: ResponseCommand, payload: bytes) -> None:
        if self._command_char_callback is None:
            return

        self._command_char_callback(None, bytearray([command.value]) + payload)

    def _get_state_char_data(self) -> bytearray:
        return pack_state(self._state)


def pack_state(state: SnoozDeviceState) -> bytearray:
    """Converts device data to device state"""
    return bytearray(
        [
            state.volume or 0x00,
            pack_bool(state.on),
            0x00,
            state.fan_speed or 0x00,
            pack_bool(state.fan_on),
            *([0] * 15),
        ]
    )


CHAR_VALUES_BY_MODEL = {
    SnoozDeviceModel.ORIGINAL: {
        MODEL_NUMBER_CHARACTERISTIC: "V2",
        FIRMWARE_REVISION_CHARACTERISTIC: "3",
        HARDWARE_REVISION_CHARACTERISTIC: "2",
    },
    SnoozDeviceModel.PRO: {
        MODEL_NUMBER_CHARACTERISTIC: "V4",
        FIRMWARE_REVISION_CHARACTERISTIC: "4.0",
        HARDWARE_REVISION_CHARACTERISTIC: "2",
        SOFTWARE_REVISION_CHARACTERISTIC: "v4.0-7-g4ba90ad2e",
    },
    SnoozDeviceModel.BREEZ: {
        MODEL_NUMBER_CHARACTERISTIC: "V4",
        FIRMWARE_REVISION_CHARACTERISTIC: "4.0",
        HARDWARE_REVISION_CHARACTERISTIC: "0",
        SOFTWARE_REVISION_CHARACTERISTIC: "v4.0-7-g4ba90ad2e",
    },
}


def pack_other_settings(state: SnoozDeviceState) -> bytearray:
    return bytearray([0] * 10) + bytearray(
        [pack_bool(state.fan_auto_enabled), state.target_temperature or 0x00]
    )


def pack_bool(value: bool | None) -> int:
    return 0x01 if value else 0x00


def unpack_bool(value: int) -> bool:
    return value == 0x01



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop, CancelledError, Future, Lock, Task
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum, IntEnum
from typing import Any, Callable

from transitions import Machine, State

from pysnooz.api import MIN_DEVICE_VOLUME, MIN_FAN_SPEED, SnoozDeviceApi
from pysnooz.const import UNEXPECTED_ERROR_LOG_MESSAGE
from pysnooz.model import SnoozDeviceCharacteristicData, SnoozDeviceState
from pysnooz.transition import Transition

_LOGGER = logging.getLogger(__name__)


class SnoozDeviceAction(IntEnum):
    GET_DEVICE_INFO = 0


@dataclass
class SnoozCommandData:
    # Noise sound properties
    on: bool | None = None
    volume: int | None = None

    # duration to transition target values
    duration: timedelta | None = None

    # Breez only properties
    fan_on: bool | None = None
    fan_speed: int | None = None
    temp_target: int | None = None
    auto_temp_enabled: bool | None = None

    action: SnoozDeviceAction | None = None

    def __repr__(self) -> str:
        operations: list[str] = []

        if self.on is not None:
            operations += ["TurnOn"] if self.on else ["TurnOff"]

        if self.fan_on is not None:
            operations += ["TurnOnFan"] if self.fan_on else ["TurnOffFan"]

        if self.volume is not None:
            operations += [f"SetVolume({self.volume}%)"]

        if self.fan_speed is not None:
            operations += [f"SetFanSpeed({self.fan_speed}%)"]

        if self.temp_target is not None:
            operations += [f"SetTargetTemperature({self.temp_target})"]

        if self.duration is not None:
            operations += [f"transition {self.duration}"]

        if self.action is SnoozDeviceAction.GET_DEVICE_INFO:
            operations += ["GetDeviceInfo"]

        if self.auto_temp_enabled is not None:
            operations += [
                f"{'Enable' if self.auto_temp_enabled else 'Disable'}AutoTemp"
            ]

        return ", ".join(operations)


def turn_on(
    volume: int | None = None, duration: timedelta | None = None
) -> SnoozCommandData:
    return SnoozCommandData(on=True, volume=volume, duration=duration)


def turn_off(duration: timedelta | None = None) -> SnoozCommandData:
    return SnoozCommandData(on=False, duration=duration)


def set_volume(volume: int) -> SnoozCommandData:
    return SnoozCommandData(volume=volume)


def get_device_info() -> SnoozCommandData:
    return SnoozCommandData(action=SnoozDeviceAction.GET_DEVICE_INFO)


# Breez only commands


def turn_fan_on(
    speed: int | None = None, duration: timedelta | None = None
) -> SnoozCommandData:
    return SnoozCommandData(fan_on=True, fan_speed=speed, duration=duration)


def turn_fan_off(duration: timedelta | None = None) -> SnoozCommandData:
    return SnoozCommandData(fan_on=False, duration=duration)


def set_fan_speed(speed: int) -> SnoozCommandData:
    return SnoozCommandData(fan_speed=speed)


def set_auto_temp_enabled(enabled: bool) -> SnoozCommandData:
    return SnoozCommandData(auto_temp_enabled=enabled)


def set_temp_target(temp: int) -> SnoozCommandData:
    return SnoozCommandData(temp_target=temp)


class SnoozCommandResultStatus(Enum):
    SUCCESSFUL = 0
    CANCELLED = 1
    DEVICE_UNAVAILABLE = 2
    UNEXPECTED_ERROR = 3


@dataclass
class SnoozCommandResult:
    status: SnoozCommandResultStatus
    duration: timedelta
    response: SnoozDeviceCharacteristicData | None = None


class CommandProcessorState(Enum):
    IDLE = 0
    EXECUTING = 1
    COMPLETE = 2


class SnoozCommandProcessor(ABC):
    def __init__(
        self,
        loop: AbstractEventLoop,
        _: Callable[[str], str],
        command: SnoozCommandData,
        start_time: datetime,
        result: Future[SnoozCommandResult],
    ) -> None:
        self.loop = loop
        self._ = _
        self.command = command
        self.start_time = start_time
        self.last_disconnect_time: datetime | None = None
        self.result = result
        self._execute_lock = Lock()
        self._result_status = SnoozCommandResultStatus.SUCCESSFUL
        self._execution_task: Task[None] | None = None
        self._total_disconnects: int = 0

        states = [
            State(CommandProcessorState.IDLE),
            State(CommandProcessorState.EXECUTING),
            State(CommandProcessorState.COMPLETE, on_enter=self._on_complete),
        ]

        not_complete = [CommandProcessorState.IDLE, CommandProcessorState.EXECUTING]

        self._machine = Machine(states=states, initial=CommandProcessorState.IDLE)

        self._machine.add_transition(
            "start_execution",
            CommandProcessorState.IDLE,
            CommandProcessorState.EXECUTING,
            before=self._before_execution_start,
        )
        self._machine.add_transition(
            "disconnected",
            not_complete,
            CommandProcessorState.IDLE,
            before=self._on_disconnect,
        )
        self._machine.add_transition(
            "cancelled",
            not_complete,
            CommandProcessorState.COMPLETE,
            before=lambda: self._abort_with_status(SnoozCommandResultStatus.CANCELLED),
        )
        self._machine.add_transition(
            "device_unavailable",
            not_complete,
            CommandProcessorState.COMPLETE,
            before=lambda: self._abort_with_status(
                SnoozCommandResultStatus.DEVICE_UNAVAILABLE
            ),
        ),
        self._machine.add_transition(
            "unhandled_exception",
            not_complete,
            CommandProcessorState.COMPLETE,
            before=lambda: self._abort_with_status(
                SnoozCommandResultStatus.UNEXPECTED_ERROR
            ),
        )
        self._machine.add_transition(
            "execution_complete",
            CommandProcessorState.EXECUTING,
            CommandProcessorState.COMPLETE,
        )

    @property
    def state(self) -> CommandProcessorState:
        return self._machine.state

    async def async_execute(
        self, api: SnoozDeviceApi, raise_on_cancel: bool = False
    ) -> None:
        async with self._execute_lock:
            self._execution_task = self.loop.create_task(
                self._async_execute_wrapper(api), name=f"Execute {self.command}"
            )

        try:
            await self._execution_task
        except CancelledError:
            if raise_on_cancel and (
                self.state == CommandProcessorState.EXECUTING
                or self._result_status == SnoozCommandResultStatus.CANCELLED
            ):
                raise

    async def _async_execute_wrapper(self, api: SnoozDeviceApi) -> None:
        try:
            # happens when a command is cancelled before execution is awaited
            if self.state == CommandProcessorState.COMPLETE:
                return

            self._machine.start_execution()

            response = await self._async_execute(api)

            # happens when a command is cancelled during execution
            if self.state != CommandProcessorState.EXECUTING:
                return

            self._machine.execution_complete(response=response)
        except Exception:
            _LOGGER.exception(
                self._(
                    f"Unexpected error while executing {self.command}\n"
                    + UNEXPECTED_ERROR_LOG_MESSAGE
                )
            )
            self.on_unhandled_exception()

    def _before_execution_start(self, **kwargs: Any) -> None:
        _LOGGER.debug(self._(f"Executing {self.command}"))

    @abstractmethod
    async def _async_execute(
        self, api: SnoozDeviceApi
    ) -> SnoozDeviceCharacteristicData | None:
        pass

    def cancel(self) -> None:
        # nothing to do if the command is already complete
        if self.state == CommandProcessorState.COMPLETE:
            return

        self._machine.cancelled()

    def on_disconnected(self) -> None:
        self._machine.disconnected()

    def on_device_unavailable(self) -> None:
        self._machine.device_unavailable()

    def on_unhandled_exception(self) -> None:
        self._machine.unhandled_exception()

    def _cancel_active_tasks(self) -> None:
        if self._execution_task is None or self._execution_task.done():
            return

        self._execution_task.cancel()
        self._execution_task = None

    def _abort_with_status(self, status: SnoozCommandResultStatus) -> None:
        self._result_status = status
        self._cancel_active_tasks()

    def _on_disconnect(self, **kwargs: Any) -> None:
        self.last_disconnect_time = datetime.now()
        self._total_disconnects += 1
        self._cancel_active_tasks()

    def _on_complete(self, **kwargs: Any) -> None:
        duration = datetime.now() - self.start_time

        message = f"Completed {self.command} ({self._result_status.name}) in {duration}"
        if self._total_disconnects > 0:
            message += f" with {self._total_disconnects} disconnects."

        response = kwargs.get("response", None)
        if response is not None:
            message += f" Response was {response}"

        _LOGGER.debug(self._(message))
        result = SnoozCommandResult(self._result_status, duration, response)
        self.result.set_result(result)


def default_log_formatter(message: str) -> str:
    return message


def create_command_processor(
    loop: AbstractEventLoop,
    start_time: datetime,
    data: SnoozCommandData,
    format_log_message: Callable[[str], str] | None = None,
) -> SnoozCommandProcessor:
    cls: type[SnoozCommandProcessor] | None = None

    if data.action is not None:
        cls = DeviceActionCommand
    elif data.duration:
        cls = TransitionedCommand
    else:
        cls = WriteDeviceStateCommand

    result = loop.create_future()

    return cls(
        loop, format_log_message or default_log_formatter, data, start_time, result
    )


class DeviceActionCommand(SnoozCommandProcessor):
    async def _async_execute(
        self, api: SnoozDeviceApi
    ) -> SnoozDeviceCharacteristicData | None:
        if self.command.action == SnoozDeviceAction.GET_DEVICE_INFO:
            return await api.async_get_info()

        raise ValueError(f"Unknown action {self.command.action}")


class WriteDeviceStateCommand(SnoozCommandProcessor):
    async def _async_execute(self, api: SnoozDeviceApi) -> None:
        if self.command.volume is not None:
            await api.async_set_volume(self.command.volume)
        if self.command.on is not None:
            await api.async_set_power(self.command.on)
        if self.command.fan_speed is not None:
            await api.async_set_fan_speed(self.command.fan_speed)
        if self.command.fan_on is not None:
            await api.async_set_fan_power(self.command.fan_on)
        if self.command.auto_temp_enabled is not None:
            await api.async_set_auto_temp_enabled(self.command.auto_temp_enabled)
        if self.command.temp_target is not None:
            await api.async_set_auto_temp_enabled(True)
            await api.async_set_auto_temp_threshold(self.command.temp_target)


class DeviceFeatureControls(ABC):
    @property
    @abstractmethod
    def min_percent(self) -> int:
        pass

    @abstractmethod
    def get_command_power(self, command: SnoozCommandData) -> bool | None:
        pass

    @abstractmethod
    def get_command_percent(self, command: SnoozCommandData) -> int | None:
        pass

    @abstractmethod
    def get_power(self, state: SnoozDeviceState) -> bool | None:
        pass

    @abstractmethod
    def get_percent(self, state: SnoozDeviceState) -> int | None:
        pass

    @abstractmethod
    async def async_set_power(self, api: SnoozDeviceApi, on: bool) -> None:
        pass

    @abstractmethod
    async def async_set_percent(self, api: SnoozDeviceApi, volume: int) -> None:
        pass


class SoundControls(DeviceFeatureControls):
    @property
    def min_percent(self) -> int:
        return MIN_DEVICE_VOLUME

    def get_command_power(self, command: SnoozCommandData) -> bool | None:
        return command.on

    def get_command_percent(self, command: SnoozCommandData) -> int | None:
        return command.volume

    def get_power(self, state: SnoozDeviceState) -> bool | None:
        return state.on

    def get_percent(self, state: SnoozDeviceState) -> int | None:
        return state.volume

    async def async_set_power(self, api: SnoozDeviceApi, on: bool) -> None:
        await api.async_set_power(on)

    async def async_set_percent(self, api: SnoozDeviceApi, volume: int) -> None:
        await api.async_set_volume(volume)


class FanControls(DeviceFeatureControls):
    @property
    def min_percent(self) -> int:
        return MIN_FAN_SPEED

    def get_command_power(self, command: SnoozCommandData) -> bool | None:
        return command.fan_on

    def get_command_percent(self, command: SnoozCommandData) -> int | None:
        return command.fan_speed

    def get_power(self, state: SnoozDeviceState) -> bool | None:
        return state.fan_on

    def get_percent(self, state: SnoozDeviceState) -> int | None:
        return state.fan_speed

    async def async_set_power(self, api: SnoozDeviceApi, on: bool) -> None:
        await api.async_set_fan_power(on)

    async def async_set_percent(self, api: SnoozDeviceApi, speed: int) -> None:
        await api.async_set_fan_speed(speed)


@dataclass
class ControlsTransition:
    feature: DeviceFeatureControls
    turning_on: bool | None
    start_percent: int
    end_percent: int

    @property
    def name(self) -> str:
        return self.feature.__class__.__name__.replace("Controls", "")


class TransitionedCommand(SnoozCommandProcessor):
    @property
    def is_resuming(self) -> bool:
        return self.last_disconnect_time is not None

    def __init__(
        self,
        loop: AbstractEventLoop,
        _: Callable[[str], str],
        data: SnoozCommandData,
        start_time: datetime,
        result: Future[SnoozCommandResult],
    ) -> None:
        if data.duration is None:
            raise ValueError("Duration must be set for transitioned commands")

        super().__init__(loop, _, data, start_time, result)
        self._transition = Transition()
        self._starting_state: SnoozDeviceState | None = None
        self._remaining_duration = data.duration
        self._features: list[DeviceFeatureControls] = []

        if data.on is not None or data.volume is not None:
            self._features.append(SoundControls())
        if data.fan_on is not None or data.fan_speed is not None:
            self._features.append(FanControls())

    async def _async_execute(self, api: SnoozDeviceApi) -> None:
        # when resuming the transition, decrease the overall duration
        # by the time disconnected to make a best effort to complete in time
        if self.last_disconnect_time is not None:
            time_since_disconnect = datetime.now() - self.last_disconnect_time
            if time_since_disconnect > self._remaining_duration:
                self._remaining_duration = timedelta(seconds=0)
            else:
                self._remaining_duration -= time_since_disconnect

        current_state = await api.async_read_state()

        if self._starting_state is None:
            self._starting_state = current_state

        # when there's no remaining duration, it means the transition
        # resumed after being disconnected for longer than the original transition,
        # so we just immediately set the target state
        if self._remaining_duration.seconds <= 0:
            for feature in self._features:
                command_power = feature.get_command_power(self.command)
                command_percent = feature.get_command_percent(self.command)

                if command_power is not None and command_power != feature.get_power(
                    current_state
                ):
                    await feature.async_set_power(api, command_power)

                if (
                    command_percent is not None
                    and command_percent != feature.get_percent(current_state)
                ):
                    await feature.async_set_percent(api, command_percent)
                elif command_percent is None and not command_power:
                    starting_percent = feature.get_percent(self._starting_state)

                    if starting_percent is not None:
                        await feature.async_set_percent(api, starting_percent)

            return

        transitions: list[ControlsTransition] = []

        for feature in self._features:
            current_power = feature.get_power(current_state)
            current_percent = feature.get_percent(current_state)
            command_percent = feature.get_command_percent(self.command)
            command_power = feature.get_command_power(self.command)

            # To prevent a delay in the transition, set it to the minimum instead of 0
            start_percent = (
                (current_percent if current_power else feature.min_percent)
                if command_power
                else current_percent
            )
            end_percent = (
                (command_percent or current_percent)
                if command_power
                else feature.min_percent
            )

            if start_percent is None:
                raise ValueError("Start percent was None")

            if end_percent is None:
                raise ValueError("End percent was None")

            # turn on the feature if necessary
            if command_power and not current_power:
                # set percent before turning on to prevent a moment with
                # the original value
                await feature.async_set_percent(api, start_percent)
                await feature.async_set_power(api, True)

            transitions.append(
                ControlsTransition(feature, command_power, start_percent, end_percent)
            )

        await self._async_transition_controls(
            api,
            transitions,
            self._remaining_duration,
        )

    async def _async_transition_controls(
        self,
        api: SnoozDeviceApi,
        controls: list[ControlsTransition],
        duration: timedelta,
    ) -> None:
        # if the device is already at the target states, avoid the transition
        if all(c.start_percent == c.end_percent for c in controls):
            return

        action = "resume" if self.is_resuming else "start"

        last_percent: dict[str, int] = {}

        for transition in controls:
            _LOGGER.debug(
                self._(
                    f"[{action}] {transition.name} {transition.start_percent}%"
                    f" to {transition.end_percent}%"
                    f"{'' if transition.turning_on else ' then turn off'} in {duration}"
                )
            )
            last_percent[transition.name] = transition.start_percent

        async def on_update(progress: float) -> None:
            nonlocal last_percent, controls

            for transition in controls:
                next_percent = int(
                    round(
                        transition.start_percent
                        + (transition.end_percent - transition.start_percent) * progress
                    )
                )

                if next_percent != last_percent.get(transition.name, None):
                    await transition.feature.async_set_percent(api, next_percent)
                    last_percent[transition.name] = next_percent

        async def on_complete() -> None:
            for transition in controls:
                if transition.turning_on:
                    continue

                nonlocal last_percent

                initial_percent: int | None = (
                    transition.feature.get_percent(self._starting_state)
                    if self._starting_state is not None
                    else None
                )

                if initial_percent is not None:
                    _LOGGER.debug(
                        self._(
                            f"[{action}] {transition.name} power off and reset to "
                            f"{initial_percent}%"
                        )
                    )
                else:
                    _LOGGER.debug(self._(f"[{action}] {transition.name} power off"))

                await transition.feature.async_set_power(api, False)

                # if we want to turn on again, make sure we reset
                # the value before starting the transition
                if initial_percent is not None:
                    await transition.feature.async_set_percent(api, initial_percent)

        await self._transition.async_run(
            self.loop, 0.0, 1.0, duration, on_update, on_complete
        )

    def _cancel_active_tasks(self) -> None:
        super()._cancel_active_tasks()
        self._transition.cancel()



from pysnooz.model import SnoozFirmwareVersion

NEW_ISSUE_URL = (
    "https://github.com/AustinBrunkhorst/pysnooz/issues/new?labels=bug"
    "&template=log-unexpected-error.yaml&title=Uncaught+exception"
)
UNEXPECTED_ERROR_LOG_MESSAGE = (
    f"1️⃣  Report this issue: {NEW_ISSUE_URL}\n"
    "2️⃣  ⬇ copy the trace and paste in the issue ⬇\n"
)

MODEL_NUMBER_CHARACTERISTIC = "00002a24-0000-1000-8000-00805f9b34fb"
FIRMWARE_REVISION_CHARACTERISTIC = "00002a26-0000-1000-8000-00805f9b34fb"
HARDWARE_REVISION_CHARACTERISTIC = "00002a27-0000-1000-8000-00805f9b34fb"
SOFTWARE_REVISION_CHARACTERISTIC = "00002a28-0000-1000-8000-00805f9b34fb"
MANUFACTURER_NAME_CHARACTERISTIC = "00002a29-0000-1000-8000-00805f9b34fb"
READ_STATE_CHARACTERISTIC = "80c37f00-cc16-11e4-8830-0800200c9a66"
WRITE_STATE_CHARACTERISTIC = "90759319-1668-44da-9ef3-492d593bd1e5"
READ_COMMAND_CHARACTERISTIC = "f0499b1b-33ab-4df8-a6f2-2484a2ad1451"


SNOOZ_ADVERTISEMENT_LENGTH = 9

FIRMWARE_PAIRING_FLAGS = 0x01
FIRMWARE_VERSION_BY_FLAGS = {
    0x04: SnoozFirmwareVersion.V2,
    0x08: SnoozFirmwareVersion.V3,
    0x0C: SnoozFirmwareVersion.V4,
    0x10: SnoozFirmwareVersion.V5,
    0x14: SnoozFirmwareVersion.V6,
    0x18: SnoozFirmwareVersion.V7,
    0x1C: SnoozFirmwareVersion.V8,
    0x20: SnoozFirmwareVersion.V9,
    0x24: SnoozFirmwareVersion.V10,
    0x28: SnoozFirmwareVersion.V11,
    0x2C: SnoozFirmwareVersion.V12,
    0x30: SnoozFirmwareVersion.V13,
    0x34: SnoozFirmwareVersion.V14,
    0x38: SnoozFirmwareVersion.V15,
}
SUPPORTED_FIRMWARE_VERSIONS = [
    SnoozFirmwareVersion.V2,
    SnoozFirmwareVersion.V3,
    SnoozFirmwareVersion.V4,
    SnoozFirmwareVersion.V5,
    SnoozFirmwareVersion.V6,
]

MODEL_NAME_SNOOZ = "Snooz"
MODEL_NAME_BREEZ = "Breez"
SUPPORTED_MODEL_NAMES = [MODEL_NAME_SNOOZ, MODEL_NAME_BREEZ]



from .advertisement import (
    SnoozAdvertisementData,
    get_device_display_name,
    parse_snooz_advertisement,
)
from .commands import (
    SnoozCommandResult,
    SnoozCommandResultStatus,
    get_device_info,
    set_auto_temp_enabled,
    set_fan_speed,
    set_temp_target,
    set_volume,
    turn_fan_off,
    turn_fan_on,
    turn_off,
    turn_on,
)
from .device import SnoozCommandData, SnoozDevice
from .model import (
    SnoozDeviceCharacteristicData,
    SnoozDeviceModel,
    SnoozDeviceState,
    SnoozFirmwareVersion,
    UnknownSnoozState,
)

__version__ = "0.9.2"

__all__ = [
    "SnoozDeviceModel",
    "SnoozFirmwareVersion",
    "SnoozDevice",
    "SnoozDeviceState",
    "SnoozDeviceCharacteristicData",
    "UnknownSnoozState",
    "SnoozCommandData",
    "SnoozCommandResult",
    "SnoozCommandResultStatus",
    "SnoozAdvertisementData",
    "parse_snooz_advertisement",
    "get_device_display_name",
    "get_device_info",
    "turn_on",
    "turn_off",
    "set_volume",
    "turn_fan_on",
    "turn_fan_off",
    "set_fan_speed",
    "set_auto_temp_enabled",
    "set_temp_target",
]
